import numpy as np
from PIL import ImageColor

hex_green = "#7db5a8"
hex_bright_green = "#83f28f"
hex_red = "#b75659"
hex_orange = "#d6ae72"
hex_gray = "#595959"
hex_blue = "#73c2fb"

rgba_green = np.array(ImageColor.getcolor(hex_green, "RGBA")) / 255
rgba_bright_green = np.array(ImageColor.getcolor(hex_bright_green, "RGBA")) / 255
rgba_red = np.array(ImageColor.getcolor(hex_red, "RGBA")) / 255
rgba_orange = np.array(ImageColor.getcolor(hex_orange, "RGBA")) / 255
rgba_gray = np.array(ImageColor.getcolor(hex_gray, "RGBA")) / 255
rgba_blue = np.array(ImageColor.getcolor(hex_blue, "RGBA")) / 255

rgba_tendon_relaxed = np.array([214, 174, 114, 190]) / 255.0
rgba_tendon_contracted = np.array([183, 86, 89, 210]) / 255.0

rgba_sand = np.array([225, 191, 146, 255]) / 255.0



from typing import List

import numpy as np
from fprs.parameters import FixedParameter
from fprs.specification import MorphologySpecification, Specification


class BrittleStarJointSpecification(Specification):
    def __init__(
        self,
        range: float,
        stiffness: float,
        damping: float,
        armature: float,
    ) -> None:
        super().__init__()
        self.stiffness = FixedParameter(value=stiffness)
        self.range = FixedParameter(value=range)
        self.damping = FixedParameter(value=damping)
        self.armature = FixedParameter(value=armature)


class BrittleStarArmSegmentSpecification(Specification):
    def __init__(
        self,
        radius: float,
        length: float,
        density: float,
        in_plane_joint_specification: BrittleStarJointSpecification,
        out_of_plane_joint_specification: BrittleStarJointSpecification,
    ) -> None:
        super().__init__()
        self.radius = FixedParameter(radius)
        self.length = FixedParameter(length)
        self.density = FixedParameter(density)
        self.in_plane_joint_specification = in_plane_joint_specification
        self.out_of_plane_joint_specification = out_of_plane_joint_specification


class BrittleStarArmSpecification(Specification):
    def __init__(
        self, segment_specifications: List[BrittleStarArmSegmentSpecification]
    ) -> None:
        super().__init__()
        self.segment_specifications = segment_specifications

    @property
    def number_of_segments(self) -> int:
        return len(self.segment_specifications)


class BrittleStarDiskSpecification(Specification):
    def __init__(self, diameter: float, height: float, density: float) -> None:
        super().__init__()
        self.radius = FixedParameter(diameter / 2)
        self.height = FixedParameter(height)
        self.density = FixedParameter(density)


class BrittleStarActuationSpecification(Specification):
    def __init__(
        self,
        use_tendons: bool,
        use_p_control: bool,
        use_torque_control: bool,
        radius_to_strength_factor: float,
    ) -> None:
        super().__init__()
        assert (
            use_p_control + use_torque_control == 1
        ), "Only one actuation method can be used."
        assert (
            not use_tendons or use_torque_control
        ), "Only torque control is supported with tendons."
        self.use_tendons = FixedParameter(use_tendons)
        self.use_p_control = FixedParameter(use_p_control)
        self.use_torque_control = FixedParameter(use_torque_control)
        self.radius_to_strength_factor = FixedParameter(radius_to_strength_factor)


class BrittleStarSensorSpecification(Specification):
    def __init__(self, num_contact_sensors_per_segment: int) -> int:
        super().__init__()
        self.num_contact_sensors_per_segment = FixedParameter(
            num_contact_sensors_per_segment
        )


class BrittleStarMorphologySpecification(MorphologySpecification):
    def __init__(
        self,
        disk_specification: BrittleStarDiskSpecification,
        arm_specifications: List[BrittleStarArmSpecification],
        actuation_specification: BrittleStarActuationSpecification,
        sensor_specification: BrittleStarSensorSpecification,
    ) -> None:
        super(BrittleStarMorphologySpecification, self).__init__()
        self.disk_specification = disk_specification
        self.arm_specifications = arm_specifications
        self.actuation_specification = actuation_specification
        self.sensor_specification = sensor_specification

    @property
    def number_of_arms(self) -> int:
        return len(self.arm_specifications)

    @property
    def number_of_non_empty_arms(self) -> int:
        return len(
            [
                number_of_segments
                for number_of_segments in self.number_of_segments_per_arm
                if number_of_segments > 0
            ]
        )

    @property
    def number_of_segments_per_arm(self) -> np.ndarray:
        return np.array(
            [
                arm_specification.number_of_segments
                for arm_specification in self.arm_specifications
            ]
        )

    @property
    def total_number_of_segments(self) -> int:
        return np.sum(self.number_of_segments_per_arm)



from typing import List, Callable

import chex
import jax.numpy as jnp
import mujoco
import numpy as np
from jax.scipy.spatial.transform import Rotation
from moojoco.environment.base import BaseObservable
from moojoco.environment.mjc_env import MJCObservable
from moojoco.environment.mjx_env import MJXObservable
from transforms3d.euler import quat2euler


def get_quat2euler_fn(backend: str) -> Callable[[chex.Array], chex.Array]:
    if backend == "mjx":

        def jquat2euler(quat):
            xywz = jnp.roll(a=quat, shift=-1)
            return Rotation.from_quat(xywz).as_euler(seq="xyz", degrees=False)

        return jquat2euler
    else:
        return quat2euler


def get_base_brittle_star_observables(
    mj_model: mujoco.MjModel, backend: str
) -> List[BaseObservable]:
    if backend == "mjx":
        observable_class = MJXObservable
        bnp = jnp
        get_data = lambda state: state.mjx_data
    else:
        observable_class = MJCObservable
        bnp = np
        get_data = lambda state: state.mj_data

    sensors = [mj_model.sensor(i) for i in range(mj_model.nsensor)]

    # All joint positions
    joint_pos_sensors = [
        sensor
        for sensor in sensors
        if sensor.type[0] == mujoco.mjtSensor.mjSENS_JOINTPOS
    ]
    joints = [mj_model.joint(sensor.objid[0]) for sensor in joint_pos_sensors]

    joint_position_observable = observable_class(
        name="joint_position",
        low=bnp.array([joint.range[0] for joint in joints]),
        high=bnp.array([joint.range[1] for joint in joints]),
        retriever=lambda state: bnp.array(
            [
                get_data(state).sensordata[
                    sensor.adr[0] : sensor.adr[0] + sensor.dim[0]
                ]
                for sensor in joint_pos_sensors
            ]
        ).flatten(),
    )

    # All joint velocities
    joint_vel_sensors = [
        sensor
        for sensor in sensors
        if sensor.type[0] == mujoco.mjtSensor.mjSENS_JOINTVEL
    ]
    joint_velocity_observable = observable_class(
        name="joint_velocity",
        low=-bnp.inf * bnp.ones(len(joint_vel_sensors)),
        high=bnp.inf * bnp.ones(len(joint_vel_sensors)),
        retriever=lambda state: bnp.array(
            [
                get_data(state).sensordata[
                    sensor.adr[0] : sensor.adr[0] + sensor.dim[0]
                ]
                for sensor in joint_vel_sensors
            ]
        ).flatten(),
    )

    # All joint actuator torques
    joint_actuator_frc_sensors = [
        sensor
        for sensor in sensors
        if sensor.type[0] == mujoco.mjtSensor.mjSENS_JOINTACTFRC
    ]
    joint_actuator_force_observable = observable_class(
        name="joint_actuator_force",
        low=-bnp.inf * bnp.ones(len(joint_actuator_frc_sensors)),
        high=bnp.inf * bnp.ones(len(joint_actuator_frc_sensors)),
        retriever=lambda state: bnp.array(
            [
                get_data(state).sensordata[
                    sensor.adr[0] : sensor.adr[0] + sensor.dim[0]
                ]
                for sensor in joint_actuator_frc_sensors
            ]
        ).flatten(),
    )

    # All actuator forces
    actuator_frc_sensors = [
        mj_model.sensor(i)
        for i in range(mj_model.nsensor)
        if mj_model.sensor(i).type[0] == mujoco.mjtSensor.mjSENS_ACTUATORFRC
    ]
    actuator_force_observable = observable_class(
        name="actuator_force",
        low=bnp.array([limits[0] for limits in mj_model.actuator_forcerange]),
        high=bnp.array([limits[1] for limits in mj_model.actuator_forcerange]),
        retriever=lambda state: bnp.array(
            [
                get_data(state).sensordata[
                    sensor.adr[0] : sensor.adr[0] + sensor.dim[0]
                ]
                for sensor in actuator_frc_sensors
            ]
        ).flatten(),
    )

    # disk pos
    disk_framepos_sensor = [
        mj_model.sensor(i)
        for i in range(mj_model.nsensor)
        if mj_model.sensor(i).type[0] == mujoco.mjtSensor.mjSENS_FRAMEPOS
        and "disk" in mj_model.sensor(i).name
    ][0]
    disk_position_observable = observable_class(
        name="disk_position",
        low=-bnp.inf * bnp.ones(3),
        high=bnp.inf * bnp.ones(3),
        retriever=lambda state: get_data(state).sensordata[
            disk_framepos_sensor.adr[0] : disk_framepos_sensor.adr[0]
            + disk_framepos_sensor.dim[0]
        ],
    )
    # disk rotation
    disk_framequat_sensor = [
        mj_model.sensor(i)
        for i in range(mj_model.nsensor)
        if mj_model.sensor(i).type[0] == mujoco.mjtSensor.mjSENS_FRAMEQUAT
        and "disk" in mj_model.sensor(i).name
    ][0]
    disk_rotation_observable = observable_class(
        name="disk_rotation",
        low=-bnp.pi * bnp.ones(3),
        high=bnp.pi * bnp.ones(3),
        retriever=lambda state: get_quat2euler_fn(backend=backend)(
            get_data(state).sensordata[
                disk_framequat_sensor.adr[0] : disk_framequat_sensor.adr[0]
                + disk_framequat_sensor.dim[0]
            ]
        ),
    )

    # disk linvel
    disk_framelinvel_sensor = [
        mj_model.sensor(i)
        for i in range(mj_model.nsensor)
        if mj_model.sensor(i).type[0] == mujoco.mjtSensor.mjSENS_FRAMELINVEL
        and "disk" in mj_model.sensor(i).name
    ][0]
    disk_linvel_observable = observable_class(
        name="disk_linear_velocity",
        low=-bnp.inf * bnp.ones(3),
        high=bnp.inf * bnp.ones(3),
        retriever=lambda state: get_data(state).sensordata[
            disk_framelinvel_sensor.adr[0] : disk_framelinvel_sensor.adr[0]
            + disk_framelinvel_sensor.dim[0]
        ],
    )

    # disk angvel
    disk_frameangvel_sensor = [
        mj_model.sensor(i)
        for i in range(mj_model.nsensor)
        if mj_model.sensor(i).type[0] == mujoco.mjtSensor.mjSENS_FRAMEANGVEL
        and "disk" in mj_model.sensor(i).name
    ][0]
    disk_angvel_observable = observable_class(
        name="disk_angular_velocity",
        low=-bnp.inf * bnp.ones(3),
        high=bnp.inf * bnp.ones(3),
        retriever=lambda state: get_data(state).sensordata[
            disk_frameangvel_sensor.adr[0] : disk_frameangvel_sensor.adr[0]
            + disk_frameangvel_sensor.dim[0]
        ],
    )

    # tendons
    tendon_pos_sensors = [
        sensor
        for sensor in sensors
        if sensor.type[0] == mujoco.mjtSensor.mjSENS_TENDONPOS
    ]
    tendon_pos_observable = observable_class(
        name="tendon_position",
        low=bnp.zeros(len(tendon_pos_sensors)),
        high=bnp.inf * bnp.ones(len(tendon_pos_sensors)),
        retriever=lambda state: bnp.array(
            [
                get_data(state).sensordata[
                    sensor.adr[0] : sensor.adr[0] + sensor.dim[0]
                ]
                for sensor in tendon_pos_sensors
            ]
        ).flatten(),
    )

    tendon_vel_sensors = [
        sensor
        for sensor in sensors
        if sensor.type[0] == mujoco.mjtSensor.mjSENS_TENDONVEL
    ]
    tendon_vel_observable = observable_class(
        name="tendon_velocity",
        low=-bnp.inf * bnp.ones(len(tendon_vel_sensors)),
        high=bnp.inf * bnp.ones(len(tendon_vel_sensors)),
        retriever=lambda state: bnp.array(
            [
                get_data(state).sensordata[
                    sensor.adr[0] : sensor.adr[0] + sensor.dim[0]
                ]
                for sensor in tendon_vel_sensors
            ]
        ).flatten(),
    )

    # contacts
    contact_sensors = [
        sensor for sensor in sensors if sensor.type[0] == mujoco.mjtSensor.mjSENS_TOUCH
    ]
    segment_contact_observable = observable_class(
        name="segment_contact",
        low=bnp.zeros(len(contact_sensors)),
        high=bnp.inf * bnp.ones(len(contact_sensors)),
        retriever=lambda state: bnp.array(
            [
                get_data(state).sensordata[
                    sensor.adr[0] : sensor.adr[0] + sensor.dim[0]
                ]
                for sensor in contact_sensors
            ]
        ).flatten(),
    )

    return [
        joint_position_observable,
        joint_velocity_observable,
        joint_actuator_force_observable,
        actuator_force_observable,
        disk_position_observable,
        disk_rotation_observable,
        disk_linvel_observable,
        disk_angvel_observable,
        tendon_pos_observable,
        tendon_vel_observable,
        segment_contact_observable,
    ]



"""
Code mainly copied from: https://github.com/pvigier/perlin-numpy/blob/master/perlin_numpy/perlin2d.py

Changes:
    - added Jax variant
    - merged jax and numpy implementation
    - added typing
    - added seed
"""

from typing import Optional, Tuple, Union

import jax
import numpy as np

ArrayType = Union[np.ndarray, jax.numpy.ndarray]


def interpolant(t: ArrayType) -> ArrayType:
    return t * t * t * (t * (t * 6 - 15) + 10)


def generate_perlin_noise_2d(
    shape: Tuple[int, int],
    res: Tuple[int, int],
    rng_key: Optional[Union[int, jax.random.PRNGKey]] = None,
    npi: Union[np, jax.numpy] = np,
) -> ArrayType:
    """Generate a 2D numpy array of perlin noise.

    Args:
        shape: The shape of the generated array (tuple of two ints).
            This must be a multple of res.
        res: The number of periods of noise to generate along each
            axis (tuple of two ints). Note: shape must be a multiple of
            res.
        rng_key: The seed used for numpy's RNG or the key for jax's RNG.
        npi: The numpy implementation to use (numpy or jax.numpy).
    Returns:
        A numpy or jax array of the given shape with the generated noise.

    Raises:
        ValueError: If shape is not a multiple of res.
    """
    if npi == jax.numpy:
        random_values = jax.random.uniform(
            rng_key, shape=(res[0] + 1, res[1] + 1), minval=0, maxval=1
        )
    else:
        rs = np.random.RandomState(seed=rng_key)
        random_values = rs.rand(res[0] + 1, res[1] + 1)

    def _generate(random_values: ArrayType) -> ArrayType:
        delta = (res[0] / shape[0], res[1] / shape[1])
        d = (shape[0] // res[0], shape[1] // res[1])
        grid = (
            npi.mgrid[0 : res[0] : delta[0], 0 : res[1] : delta[1]].transpose(1, 2, 0)
            % 1
        )
        # Gradients
        angles = 2 * npi.pi * random_values
        gradients = npi.dstack((npi.cos(angles), npi.sin(angles)))
        gradients = gradients.repeat(d[0], 0).repeat(d[1], 1)
        g00 = gradients[: -d[0], : -d[1]]
        g10 = gradients[d[0] :, : -d[1]]
        g01 = gradients[: -d[0], d[1] :]
        g11 = gradients[d[0] :, d[1] :]
        # Ramps
        n00 = npi.sum(npi.dstack((grid[:, :, 0], grid[:, :, 1])) * g00, 2)
        n10 = npi.sum(npi.dstack((grid[:, :, 0] - 1, grid[:, :, 1])) * g10, 2)
        n01 = npi.sum(npi.dstack((grid[:, :, 0], grid[:, :, 1] - 1)) * g01, 2)
        n11 = npi.sum(npi.dstack((grid[:, :, 0] - 1, grid[:, :, 1] - 1)) * g11, 2)
        # Interpolation
        t = interpolant(grid)
        n0 = n00 * (1 - t[:, :, 0]) + t[:, :, 0] * n10
        n1 = n01 * (1 - t[:, :, 0]) + t[:, :, 0] * n11
        noise = npi.sqrt(2) * ((1 - t[:, :, 1]) * n0 + t[:, :, 1] * n1)

        return (noise - np.min(noise)) / (np.max(noise) - np.min(noise))

    if npi == jax.numpy:
        return jax.jit(_generate)(random_values)
    else:
        return _generate(random_values)



from typing import Tuple

import numpy as np
from moojoco.mjcf.arena import ArenaConfiguration, MJCFArena

from biorobot.brittle_star.mjcf.arena.target import MJCFTarget
from biorobot.utils import colors


class AquariumArenaConfiguration(ArenaConfiguration):
    def __init__(
        self,
        name: str = "AquariumArena",
        size: Tuple[int, int] = (5, 3),
        sand_ground_color: bool = False,
        light_map_resolution: int = 50,
        attach_target: bool = False,
        wall_height: float = 1.5,
        wall_thickness: float = 0.1,
    ) -> None:
        super().__init__(name=name)
        self.size = size
        self.sand_ground_color = sand_ground_color
        self.light_map_resolution = light_map_resolution
        self.attach_target = attach_target
        self.wall_height = wall_height
        self.wall_thickness = wall_thickness


class MJCFAquariumArena(MJCFArena):
    @property
    def arena_configuration(self) -> AquariumArenaConfiguration:
        return super().arena_configuration

    def _build(self, *args, **kwargs) -> None:
        self._configure_cameras()
        self._configure_lights()
        self._configure_water()
        self._configure_sky()
        self._build_ground()
        self._build_walls()
        self._build_target()

    def _configure_cameras(self) -> None:
        self.mjcf_model.worldbody.add(
            "camera", name="top_camera", pos=[0, 0, 20], quat=[1, 0, 0, 0]
        )

    def _configure_lights(self) -> None:
        self.mjcf_model.worldbody.add(
            "light",
            pos=[-20, 0, 20],
            directional=True,
            dir=[0, 0, -0.5],
            diffuse=[0.1, 0.1, 0.1],
            castshadow=False,
        )
        self.mjcf_model.visual.headlight.set_attributes(
            ambient=[0.4, 0.4, 0.4], diffuse=[0.8, 0.8, 0.8], specular=[0.1, 0.1, 0.1]
        )

    def _configure_water(self) -> None:
        self.mjcf_model.option.density = 1000
        self.mjcf_model.option.viscosity = 0.0009

    def _configure_sky(self) -> None:
        # white sky
        self.mjcf_model.asset.add(
            "texture",
            type="skybox",
            builtin="flat",
            rgb1="1.0 1.0 1.0",
            rgb2="1.0 1.0 1.0",
            width=200,
            height=200,
        )

    def _build_ground(self) -> None:
        if self.arena_configuration.sand_ground_color:
            w, h = self.arena_configuration.size
            resolution = self.arena_configuration.light_map_resolution
            ground_texture = self.mjcf_model.asset.add(
                "texture",
                type="2d",
                builtin="flat",
                name="groundplane",
                rgb1=colors.rgba_sand[:3],
                width=resolution * w,
                height=resolution * h,
            )
            ground_material = self.mjcf_model.asset.add(
                "material",
                name="groundplane",
                reflectance=0.0,
                texture=ground_texture,
            )
            rgba = colors.rgba_sand
        else:
            ground_texture = self.mjcf_model.asset.add(
                "texture",
                rgb1=[0.2, 0.3, 0.4],
                rgb2=[0.1, 0.2, 0.3],
                type="2d",
                builtin="checker",
                name="groundplane",
                width=200,
                height=200,
                mark="edge",
                markrgb=[0.8, 0.8, 0.8],
            )
            ground_material = self.mjcf_model.asset.add(
                "material",
                name="groundplane",
                texrepeat=[2, 2],  # Makes white squares exactly 1x1 length units.
                texuniform=True,
                reflectance=0.2,
                texture=ground_texture,
            )
            rgba = None

        # Build groundplane.
        self._ground_geom = self.mjcf_body.add(
            "geom",
            type="plane",
            name="groundplane",
            material=ground_material,
            rgba=rgba,
            size=list(self.arena_configuration.size) + [0.25],
            condim=4,
            contype=0,
            conaffinity=1,
        )

    def _build_walls(self) -> None:
        wall_rgba = np.asarray([115, 147, 179, 50]) / 255

        walls = []
        walls.append(
            self.mjcf_model.worldbody.add(
                "geom",
                type="box",
                name="north_wall",
                size=[
                    self.arena_configuration.size[0],
                    self.arena_configuration.wall_thickness,
                    self.arena_configuration.wall_height,
                ],
                pos=[
                    0.0,
                    self.arena_configuration.size[1]
                    + self.arena_configuration.wall_thickness,
                    self.arena_configuration.wall_height,
                ],
                rgba=wall_rgba,
                contype=0,
                conaffinity=0,
            )
        )
        walls.append(
            self.mjcf_model.worldbody.add(
                "geom",
                type="box",
                name="south_wall",
                size=[
                    self.arena_configuration.size[0],
                    self.arena_configuration.wall_thickness,
                    self.arena_configuration.wall_height,
                ],
                pos=[
                    0.0,
                    -self.arena_configuration.size[1]
                    - self.arena_configuration.wall_thickness,
                    self.arena_configuration.wall_height,
                ],
                rgba=wall_rgba,
                contype=0,
                conaffinity=0,
            )
        )
        walls.append(
            self.mjcf_model.worldbody.add(
                "geom",
                type="box",
                name="east_wall",
                size=[
                    self.arena_configuration.wall_thickness,
                    self.arena_configuration.size[1]
                    + 2 * self.arena_configuration.wall_thickness,
                    self.arena_configuration.wall_height,
                ],
                pos=[
                    -self.arena_configuration.size[0]
                    - self.arena_configuration.wall_thickness,
                    0.0,
                    self.arena_configuration.wall_height,
                ],
                rgba=wall_rgba,
                contype=0,
                conaffinity=0,
            )
        )
        walls.append(
            self.mjcf_model.worldbody.add(
                "geom",
                type="box",
                name="west_wall",
                size=[
                    self.arena_configuration.wall_thickness,
                    self.arena_configuration.size[1]
                    + 2 * self.arena_configuration.wall_thickness,
                    self.arena_configuration.wall_height,
                ],
                pos=[
                    self.arena_configuration.size[0]
                    + self.arena_configuration.wall_thickness,
                    0.0,
                    self.arena_configuration.wall_height,
                ],
                rgba=wall_rgba,
                contype=0,
                conaffinity=0,
            )
        )

        eulers = [
            [np.pi / 2, 0, 0],
            [-np.pi / 2, 0, 0],
            [0, np.pi / 2, 0],
            [0, -np.pi / 2, 0],
        ]
        for wall, euler in zip(walls, eulers):
            self.mjcf_model.worldbody.add(
                "geom",
                type="plane",
                name=wall.name + "_collider",
                pos=wall.pos,
                size=wall.size,
                euler=euler,
                rgba=[0, 0, 0, 0],
                contype=0,
                conaffinity=1,
            )

    def _build_target(self) -> None:
        if self.arena_configuration.attach_target:
            self._target = MJCFTarget(parent=self, name="target")


if __name__ == "__main__":
    MJCFAquariumArena(
        AquariumArenaConfiguration(sand_ground_color=True)
    ).export_to_xml_with_assets("./mjcf")



from typing import Dict

import chex
import jax.random
import numpy as np
from gymnasium.envs.tabular.blackjack import EnvState
from moojoco.environment.mjx_spaces import Box

from biorobot.brittle_star.environment.undirected_locomotion.dual import (
    BrittleStarUndirectedLocomotionEnvironment,
)
from biorobot.brittle_star.environment.undirected_locomotion.shared import (
    BrittleStarUndirectedLocomotionEnvironmentConfiguration,
)
from biorobot.brittle_star.mjcf.arena.aquarium import (
    AquariumArenaConfiguration,
    MJCFAquariumArena,
)
from biorobot.brittle_star.mjcf.morphology.morphology import MJCFBrittleStarMorphology
from biorobot.brittle_star.mjcf.morphology.specification.default import (
    default_brittle_star_morphology_specification,
)


def create_env(
    backend: str, render_mode: str
) -> BrittleStarUndirectedLocomotionEnvironment:
    NUM_SEGMENTS_PER_ARM = 3
    morphology_spec = default_brittle_star_morphology_specification(
        num_arms=5,
        num_segments_per_arm=[0, NUM_SEGMENTS_PER_ARM, 0, 0, NUM_SEGMENTS_PER_ARM],
        use_p_control=True,
        use_torque_control=False,
        radius_to_strength_factor=200,
    )
    morphology = MJCFBrittleStarMorphology(morphology_spec)
    arena_config = AquariumArenaConfiguration()
    arena = MJCFAquariumArena(configuration=arena_config)

    env_config = BrittleStarUndirectedLocomotionEnvironmentConfiguration(
        render_mode=render_mode,
        num_physics_steps_per_control_step=10,
        simulation_time=5,
        time_scale=1,
        camera_ids=[0, 1],
        color_contacts=True,
        solver_iterations=1,
        solver_ls_iterations=5,
    )
    env = BrittleStarUndirectedLocomotionEnvironment.from_morphology_and_arena(
        morphology=morphology, arena=arena, configuration=env_config, backend=backend
    )
    return env


def generate_rowing_action(
    action_space: Box, observation_space: Dict, state: EnvState
) -> chex.Array:
    """
    Use P-control to generate oval motion with the bases of the arms (up to half of segments)
    Set force of remaining segments to zero
    """

    NUM_ACTUATED_SEGMENTS = 3
    NUM_ARMS = 2

    time = state.info["time"]

    try:
        ip_rom = observation_space.spaces["joint_position"].high[0]
        oop_rom = observation_space.spaces["joint_position"].high[1]
    except Exception:
        ip_rom = observation_space["joint_position"].high[0]
        oop_rom = observation_space["joint_position"].high[1]

    frequency = 3
    ip_pos = ip_rom * np.cos(frequency * time)
    oop_pos = 0.4 * oop_rom * np.sin(frequency * time)

    positions_per_arm = np.zeros_like(action_space.low).reshape((NUM_ARMS, -1))
    positions_per_arm[0, ::2] = ip_pos
    positions_per_arm[1, ::2] = -ip_pos
    positions_per_arm[:, 1::2] = oop_pos

    return positions_per_arm.flatten()

    target_positions = positions_per_arm.flatten()
    current_positions = state.observations["joint_position"]

    forces = 10 * (target_positions - current_positions)

    # Only actuate targeted segments
    forces_per_arm = forces.reshape((NUM_ARMS, -1))
    forces_per_arm[:, NUM_ACTUATED_SEGMENTS * 2 :] = 0
    forces = forces_per_arm.flatten()

    return forces

    # Only actuate targeted segments
    # forces_per_arm[:, NUM_ACTUATED_SEGMENTS * 2:] = 0
    #
    # forces = forces_per_arm.flatten()
    #
    # Rescale
    # forces = forces * (action_space.high - action_space.low) - action_space.low

    return positions_per_arm.flatten()


if __name__ == "__main__":
    BACKEND = "MJC"
    RENDER_MODE = "human"

    env = create_env(backend=BACKEND, render_mode=RENDER_MODE)

    if BACKEND == "MJC":
        env_rng = np.random.RandomState(0)
        step_fn = env.step
        reset_fn = env.reset
    else:
        env_rng = jax.random.PRNGKey(0)

        step_fn = jax.jit(env.step)
        reset_fn = jax.jit(env.reset)

    state = reset_fn(env_rng)
    while True:
        action = generate_rowing_action(env.action_space, env.observation_space, state)
        state = step_fn(state=state, action=action)
        env.render(state=state)
        if state.truncated or state.terminated:
            state = reset_fn(env_rng)

    env.close()



from __future__ import annotations

from typing import Any, Dict, List, Tuple

import mujoco
import numpy as np
from moojoco.environment.mjc_env import MJCEnv, MJCEnvState, MJCObservable
from transforms3d.euler import euler2quat

from biorobot.brittle_star.environment.directed_locomotion.shared import (
    BrittleStarDirectedLocomotionEnvironmentBase,
    BrittleStarDirectedLocomotionEnvironmentConfiguration,
)
from biorobot.brittle_star.environment.shared.observables import (
    get_base_brittle_star_observables,
)
from biorobot.brittle_star.mjcf.arena.aquarium import MJCFAquariumArena
from biorobot.brittle_star.mjcf.morphology.morphology import MJCFBrittleStarMorphology


class BrittleStarDirectedLocomotionMJCEnvironment(
    BrittleStarDirectedLocomotionEnvironmentBase, MJCEnv
):
    metadata = {"render_modes": ["human", "rgb_array"]}

    def __init__(
        self,
        mjcf_str: str,
        mjcf_assets: Dict[str, Any],
        configuration: BrittleStarDirectedLocomotionEnvironmentConfiguration,
    ) -> None:
        BrittleStarDirectedLocomotionEnvironmentBase.__init__(self)
        MJCEnv.__init__(
            self,
            mjcf_str=mjcf_str,
            mjcf_assets=mjcf_assets,
            configuration=configuration,
        )
        self._cache_references(mj_model=self.frozen_mj_model)

    @property
    def environment_configuration(
        self,
    ) -> BrittleStarDirectedLocomotionEnvironmentConfiguration:
        return super(MJCEnv, self).environment_configuration

    @classmethod
    def from_morphology_and_arena(
        cls,
        morphology: MJCFBrittleStarMorphology,
        arena: MJCFAquariumArena,
        configuration: BrittleStarDirectedLocomotionEnvironmentConfiguration,
    ) -> BrittleStarDirectedLocomotionMJCEnvironment:
        assert arena.arena_configuration.attach_target, (
            f"Arena must have a target attached. Please set "
            f"'attach_target' to 'True' in the "
            f"AquariumArenaConfiguration."
        )
        return super().from_morphology_and_arena(
            morphology=morphology, arena=arena, configuration=configuration
        )

    @staticmethod
    def _get_xy_direction_to_target(state: MJCEnvState) -> np.ndarray:
        target_position = state.mj_data.body("target").xpos
        disk_position = state.mj_data.body("BrittleStarMorphology/central_disk").xpos
        direction_to_target = target_position - disk_position
        return direction_to_target[:2]

    @staticmethod
    def _get_xy_distance_to_target(state: MJCEnvState) -> float:
        xy_direction_to_target = (
            BrittleStarDirectedLocomotionMJCEnvironment._get_xy_direction_to_target(
                state=state
            )
        )
        xy_distance_to_target = np.linalg.norm(xy_direction_to_target)
        return xy_distance_to_target

    @staticmethod
    def _get_xy_target_position(state: MJCEnvState) -> np.ndarray:
        return np.array(state.mj_data.body("target").xpos[:2])

    def _create_observables(self) -> List[MJCObservable]:
        base_observables = get_base_brittle_star_observables(
            mj_model=self.frozen_mj_model, backend="mjc"
        )

        # direction to target
        unit_xy_direction_to_target_observable = MJCObservable(
            name="unit_xy_direction_to_target",
            low=-np.ones(2),
            high=np.ones(2),
            retriever=lambda state: self._get_xy_direction_to_target(state)
            / self._get_xy_distance_to_target(state=state),
        )

        # distance to target
        xy_distance_to_target_observable = MJCObservable(
            name="xy_distance_to_target",
            low=np.zeros(1),
            high=np.inf * np.ones(1),
            retriever=lambda state: np.array(
                [self._get_xy_distance_to_target(state=state)]
            ),
        )

        return base_observables + [
            unit_xy_direction_to_target_observable,
            xy_distance_to_target_observable,
        ]

    @staticmethod
    def _get_time(state: MJCEnvState) -> float:
        return state.mj_data.time

    def _get_mj_models_and_datas_to_render(
        self, state: MJCEnvState
    ) -> Tuple[List[mujoco.MjModel], List[mujoco.MjData]]:
        mj_models, mj_datas = super()._get_mj_models_and_datas_to_render(state=state)
        if self.environment_configuration.color_contacts:
            self._color_segment_capsule_contacts(
                mj_models=mj_models, contacts=state.observations["segment_contact"]
            )
        return mj_models, mj_datas

    def _get_target_position(
        self, rng: np.random.RandomState, target_position: np.ndarray | None = None
    ) -> np.ndarray:
        if target_position is not None:
            position = np.array(target_position)
        else:
            angle = rng.uniform(0, 2 * np.pi)
            radius = self.environment_configuration.target_distance
            position = np.array([radius * np.cos(angle), radius * np.sin(angle), 0.05])
        return position

    def reset(
        self,
        rng: np.random.RandomState,
        target_position: np.ndarray | None = None,
        *args,
        **kwargs,
    ) -> MJCEnvState:
        mj_model, mj_data = self._prepare_reset()

        # Set random target position
        mj_model.body("target").pos = self._get_target_position(
            rng=rng, target_position=target_position
        )

        # Set morphology position
        mj_model.body("BrittleStarMorphology/central_disk").pos[2] = 0.11

        morphology_qpos_adr = mj_model.joint(
            "BrittleStarMorphology/freejoint/"
        ).qposadr[0]

        if self.environment_configuration.random_initial_rotation:
            z_axis_rotation = rng.uniform(-np.pi, np.pi)
            quat = euler2quat(0, 0, z_axis_rotation, axes="sxyz")
            mj_data.qpos[morphology_qpos_adr + 3 : morphology_qpos_adr + 7] = quat

        # Add noise to initial qpos and qvel of segment joints
        joint_qpos_adrs = self._get_segment_joints_qpos_adrs(mj_model=mj_model)
        joint_qvel_adrs = self._get_segment_joints_qvel_adrs(mj_model=mj_model)
        num_segment_joints = len(joint_qpos_adrs)

        mj_data.qpos[joint_qpos_adrs] = mj_model.qpos0[joint_qpos_adrs] + rng.uniform(
            low=-self.environment_configuration.joint_randomization_noise_scale,
            high=self.environment_configuration.joint_randomization_noise_scale,
            size=num_segment_joints,
        )
        mj_data.qvel[joint_qvel_adrs] = rng.uniform(
            low=-self.environment_configuration.joint_randomization_noise_scale,
            high=self.environment_configuration.joint_randomization_noise_scale,
            size=num_segment_joints,
        )

        state = self._finish_reset(models_and_datas=(mj_model, mj_data), rng=rng)
        return state



import matplotlib.pyplot as plt
import numpy as np

if __name__ == "__main__":
    t = np.linspace(0, 5, 100)

    frequency = 1
    phase_offset = 0

    x = np.cos(frequency * t + phase_offset)
    y = 0.5 * np.sin(frequency * t + phase_offset)

    # Create a color gradient from light blue to dark blue
    colors = plt.cm.Blues(np.linspace(0.3, 1, len(t)))  # Using the 'Blues' colormap

    plt.scatter(x, y, c=colors, edgecolor="k")  # Add edgecolor for visibility
    plt.show()
    plt.close()



from functools import partial
from typing import Tuple

import chex
import jax.numpy as jnp
import jax.random
import numpy as np
from moojoco.environment.mjc_env import ThreadedVectorMJCEnvWrapper

from biorobot.brittle_star.usage_examples.directed_locomotion_single import post_render
from biorobot.brittle_star.usage_examples.light_escape_single import create_env

if __name__ == "__main__":
    BACKEND = "MJX"
    RENDER_MODE = "rgb_array"
    NUM_ENVS = 2

    if BACKEND == "MJC":
        env_rng, action_rng = [np.random.RandomState(i) for i in range(NUM_ENVS)], None

        env = ThreadedVectorMJCEnvWrapper(
            create_env_fn=partial(create_env, backend=BACKEND, render_mode=RENDER_MODE),
            num_environments=NUM_ENVS,
        )

        step_fn = env.step
        reset_fn = env.reset

        def action_sample_fn(_: None) -> Tuple[np.ndarray, None]:
            return env.action_space.sample(), None

    else:
        action_rng, env_rng = jax.random.split(jax.random.PRNGKey(0), 2)
        action_rng = jnp.array(jax.random.split(action_rng, NUM_ENVS))
        env_rng = jnp.array(jax.random.split(env_rng, NUM_ENVS))

        env = create_env(backend=BACKEND, render_mode=RENDER_MODE)

        step_fn = jax.jit(jax.vmap(env.step))
        reset_fn = jax.jit(jax.vmap(env.reset))

        def action_sample_fn(rng: chex.PRNGKey) -> Tuple[jnp.ndarray, chex.PRNGKey]:
            rng, sub_rng = jax.random.split(rng, 2)
            return env.action_space.sample(rng=sub_rng), rng

        action_sample_fn = jax.jit(jax.vmap(action_sample_fn))

    state = reset_fn(env_rng)

    while True:
        action, action_rng = action_sample_fn(action_rng)
        state = step_fn(state=state, action=action)
        post_render(
            env.render(state=state),
            environment_configuration=env.environment_configuration,
        )
    env.close()



from moojoco.environment.dual import DualMuJoCoEnvironment

from biorobot.brittle_star.environment.undirected_locomotion.mjc_env import (
    BrittleStarUndirectedLocomotionMJCEnvironment,
)
from biorobot.brittle_star.environment.undirected_locomotion.mjx_env import (
    BrittleStarUndirectedLocomotionMJXEnvironment,
)


class BrittleStarUndirectedLocomotionEnvironment(DualMuJoCoEnvironment):
    MJC_ENV_CLASS = BrittleStarUndirectedLocomotionMJCEnvironment
    MJX_ENV_CLASS = BrittleStarUndirectedLocomotionMJXEnvironment

    def __init__(
        self,
        env: (
            BrittleStarUndirectedLocomotionMJCEnvironment
            | BrittleStarUndirectedLocomotionMJXEnvironment
        ),
        backend: str,
    ) -> None:
        super().__init__(env=env, backend=backend)



from moojoco.environment.dual import DualMuJoCoEnvironment

from biorobot.brittle_star.environment.directed_locomotion.mjc_env import (
    BrittleStarDirectedLocomotionMJCEnvironment,
)
from biorobot.brittle_star.environment.directed_locomotion.mjx_env import (
    BrittleStarDirectedLocomotionMJXEnvironment,
)


class BrittleStarDirectedLocomotionEnvironment(DualMuJoCoEnvironment):
    MJC_ENV_CLASS = BrittleStarDirectedLocomotionMJCEnvironment
    MJX_ENV_CLASS = BrittleStarDirectedLocomotionMJXEnvironment

    def __init__(
        self,
        env: (
            BrittleStarDirectedLocomotionMJCEnvironment
            | BrittleStarDirectedLocomotionMJXEnvironment
        ),
        backend: str,
    ) -> None:
        super().__init__(env=env, backend=backend)



import abc
from typing import List, Union

import chex
import mujoco
import numpy as np
from moojoco.environment.base import BaseEnvState
from moojoco.environment.renderer import MujocoRenderer

from biorobot.brittle_star.environment.shared.base import (
    BrittleStarEnvironmentBase,
    BrittleStarEnvironmentBaseConfiguration,
)
from biorobot.utils import colors


class BrittleStarLightEscapeEnvironmentConfiguration(
    BrittleStarEnvironmentBaseConfiguration
):
    def __init__(
        self,
        light_perlin_noise_scale: int = 0,
        *args,
        **kwargs,
    ) -> None:
        assert light_perlin_noise_scale == 0 or 200 % light_perlin_noise_scale == 0, (
            "Please only provide integer factors of 200 for the "
            "'light_perlin_noise_scale' parameter."
        )
        super().__init__(
            *args,
            **kwargs,
        )
        self.light_perlin_noise_scale = int(light_perlin_noise_scale)


class BrittleStarLightEscapeEnvironmentBase(BrittleStarEnvironmentBase):
    def __init__(self) -> None:
        super().__init__()

        self._segment_capsule_areas: chex.Array | None = None
        self._disk_area: float | None = None

    def _update_reward(
        self, state: BaseEnvState, previous_state: BaseEnvState
    ) -> BaseEnvState:
        previous_light_income = self._get_body_light_income(state=previous_state)
        current_light_income = self._get_body_light_income(state=state)
        reward = previous_light_income - current_light_income

        # noinspection PyUnresolvedReferences
        return state.replace(reward=reward)

    def _get_body_light_income(self, state: BaseEnvState) -> float:
        segment_light_values = self._get_light_per_segment(state=state)
        disk_light_value = self._get_disk_light_income(state=state)

        segment_light_values = (
            segment_light_values * self._segment_capsule_areas
        ).sum()
        disk_light_value = disk_light_value * self._disk_area

        body_light = (segment_light_values + disk_light_value) / (
            self._segment_capsule_areas.sum() + self._disk_area
        )
        return body_light

    def _update_info(self, state: BaseEnvState) -> BaseEnvState:
        info = state.info
        info.update(
            {
                "time": self._get_time(state=state),
                "x_distance_travelled": self._get_x_distance_from_start_position(
                    state=state
                ),
            }
        )
        # noinspection PyUnresolvedReferences
        return state.replace(info=info)

    def _update_terminated(self, state: BaseEnvState) -> BaseEnvState:
        # noinspection PyUnresolvedReferences
        return state.replace(terminated=False)

    def _update_truncated(self, state: BaseEnvState) -> BaseEnvState:
        truncated = (
            self._get_time(state=state) > self.environment_configuration.simulation_time
        )

        # noinspection PyUnresolvedReferences
        return state.replace(truncated=truncated)

    @staticmethod
    def _update_mj_models_tex_data(
        mj_models: List[mujoco.MjModel], state: BaseEnvState
    ) -> None:
        if np.any(state.info["_light_map_has_changed"]):
            ground_texture = state.mj_model.texture("groundplane")
            h, w, adr = (
                ground_texture.height[0],
                ground_texture.width[0],
                ground_texture.adr[0],
            )
            size = h * w * 3
            # Update textures in mj models
            light_maps = state.info["_light_map"]
            if len(mj_models) == 1:
                light_maps = [light_maps]

            for mj_model, light_map in zip(mj_models, light_maps):
                coloured_light_map = 0.3 + 0.7 * np.array(light_map)
                coloured_light_map = np.stack((coloured_light_map,) * 3, axis=-1)
                coloured_light_map = coloured_light_map * colors.rgba_sand[:3]
                mj_model.tex_data[adr : adr + size] = coloured_light_map.flatten() * 255

    def _update_renderer_context(
        self,
        mj_model: mujoco.MjModel,
        state: BaseEnvState,
        renderer: Union[MujocoRenderer, mujoco.Renderer],
    ) -> None:
        if np.any(state.info["_light_map_has_changed"]):
            ground_texture = mj_model.texture("groundplane")
            context = self.get_renderer_context(renderer=renderer)
            mujoco.mjr_uploadTexture(m=mj_model, con=context, texid=ground_texture.id)

    @staticmethod
    def _get_x_start_position(mj_model: mujoco.MjModel) -> float:
        arena_size = mj_model.geom("groundplane").size[0]
        return -0.75 * arena_size

    @property
    @abc.abstractmethod
    def environment_configuration(
        self,
    ) -> BrittleStarLightEscapeEnvironmentConfiguration:
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def _get_time(state: BaseEnvState) -> float:
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def _get_x_distance_from_start_position(state: BaseEnvState) -> float:
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def _get_light_value_at_xy_positions(
        state: BaseEnvState, xy_positions: chex.Array
    ) -> float:
        raise NotImplementedError

    @abc.abstractmethod
    def _get_light_per_segment(self, state: BaseEnvState) -> chex.Array:
        raise NotImplementedError

    @abc.abstractmethod
    def _get_disk_light_income(self, state: BaseEnvState) -> float:
        raise NotImplementedError

    @abc.abstractmethod
    def get_renderer_context(
        self, renderer: Union[MujocoRenderer, mujoco.Renderer]
    ) -> mujoco.MjrContext:
        raise NotImplementedError



import atexit
import shutil
from pathlib import Path

brb_tmp_directory = ""


def set_brb_tmp_directory(new_path: str) -> None:
    global brb_tmp_directory
    brb_tmp_directory = new_path
    Path(brb_tmp_directory).mkdir(parents=True, exist_ok=True)


set_brb_tmp_directory("/tmp/brb")


def exit_handler():
    global brb_tmp_directory
    if Path(brb_tmp_directory).exists():
        shutil.rmtree(brb_tmp_directory, ignore_errors=True)


atexit.register(exit_handler)



import abc

from moojoco.environment.base import BaseEnvState

from biorobot.brittle_star.environment.shared.base import (
    BrittleStarEnvironmentBase,
    BrittleStarEnvironmentBaseConfiguration,
)


class BrittleStarUndirectedLocomotionEnvironmentConfiguration(
    BrittleStarEnvironmentBaseConfiguration
):
    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(
            *args,
            **kwargs,
        )


class BrittleStarUndirectedLocomotionEnvironmentBase(BrittleStarEnvironmentBase):
    def __init__(self) -> None:
        super().__init__()

    @staticmethod
    def _update_terminated(state: BaseEnvState) -> BaseEnvState:
        # noinspection PyUnresolvedReferences
        return state.replace(terminated=False)

    def _update_truncated(self, state: BaseEnvState) -> BaseEnvState:
        truncated = (
            self._get_time(state=state) > self.environment_configuration.simulation_time
        )
        # noinspection PyUnresolvedReferences
        return state.replace(truncated=truncated)

    def _update_reward(
        self, state: BaseEnvState, previous_state: BaseEnvState
    ) -> BaseEnvState:
        previous_distance_from_origin = self._get_xy_distance_from_origin(
            state=previous_state
        )
        current_distance_from_origin = self._get_xy_distance_from_origin(state=state)
        reward = current_distance_from_origin - previous_distance_from_origin

        # noinspection PyUnresolvedReferences
        return state.replace(reward=reward)

    def _update_info(self, state: BaseEnvState) -> BaseEnvState:
        info = state.info
        info.update(
            {
                "time": self._get_time(state=state),
                "xy_distance_from_origin": self._get_xy_distance_from_origin(
                    state=state
                ),
            }
        )

        # noinspection PyUnresolvedReferences
        return state.replace(info=info)

    @staticmethod
    @abc.abstractmethod
    def _get_time(state: BaseEnvState) -> float:
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def _get_xy_distance_from_origin(state: BaseEnvState) -> float:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def environment_configuration(
        self,
    ) -> BrittleStarUndirectedLocomotionEnvironmentConfiguration:
        raise NotImplementedError



import jax.numpy as jnp
import jax.random

from biorobot.brittle_star.environment.directed_locomotion.dual import (
    BrittleStarDirectedLocomotionEnvironment,
)
from biorobot.brittle_star.environment.directed_locomotion.shared import (
    BrittleStarDirectedLocomotionEnvironmentConfiguration,
)
from biorobot.brittle_star.environment.light_escape.dual import (
    BrittleStarLightEscapeEnvironment,
)
from biorobot.brittle_star.environment.light_escape.shared import (
    BrittleStarLightEscapeEnvironmentConfiguration,
)
from biorobot.brittle_star.mjcf.arena.aquarium import (
    AquariumArenaConfiguration,
    MJCFAquariumArena,
)
from biorobot.brittle_star.mjcf.morphology.morphology import MJCFBrittleStarMorphology
from biorobot.brittle_star.mjcf.morphology.specification.default import (
    default_brittle_star_morphology_specification,
)


def create_morphology() -> MJCFBrittleStarMorphology:
    morphology_spec = default_brittle_star_morphology_specification(
        num_arms=5,
        num_segments_per_arm=6,
        use_p_control=False,
        use_torque_control=True,
        radius_to_strength_factor=400,
    )
    return MJCFBrittleStarMorphology(morphology_spec)


def create_dirlo_env(
    backend: str, render_mode: str
) -> BrittleStarDirectedLocomotionEnvironment:
    morphology = create_morphology()
    arena_config = AquariumArenaConfiguration(attach_target=True)
    arena = MJCFAquariumArena(configuration=arena_config)
    env_config = BrittleStarDirectedLocomotionEnvironmentConfiguration(
        render_mode=render_mode,
        num_physics_steps_per_control_step=10,
        simulation_time=5,
        time_scale=2,
        camera_ids=[0, 1],
        color_contacts=True,
    )
    env = BrittleStarDirectedLocomotionEnvironment.from_morphology_and_arena(
        morphology=morphology, arena=arena, configuration=env_config, backend=backend
    )
    return env


def create_light_escape_env(
    backend: str, render_mode: str
) -> BrittleStarLightEscapeEnvironment:
    morphology = create_morphology()
    arena_config = AquariumArenaConfiguration(sand_ground_color=True)
    arena = MJCFAquariumArena(configuration=arena_config)

    env_config = BrittleStarLightEscapeEnvironmentConfiguration(
        render_mode=render_mode,
        light_perlin_noise_scale=10,
        num_physics_steps_per_control_step=10,
        simulation_time=5,
        time_scale=2,
        camera_ids=[0, 1],
        color_contacts=True,
    )
    env = BrittleStarLightEscapeEnvironment.from_morphology_and_arena(
        morphology=morphology, arena=arena, configuration=env_config, backend=backend
    )
    return env


def clip_and_rescale(
    values: jax.Array,
    original_low: jax.Array,
    original_high: jax.Array,
    new_low: jax.Array,
    new_high: jax.Array,
) -> jax.Array:
    clipped = jnp.clip(values, original_low, original_high)
    # Normalizing to 0-1
    normalized = (clipped - original_low) / (original_high - original_low)

    # Scaling to new range
    scaled = normalized * (new_high - new_low) + new_low

    return scaled


if __name__ == "__main__":
    # Create env
    env = create_dirlo_env("MJX", "human")

    step_fn = jax.jit(env.step)
    reset_fn = jax.jit(env.reset)

    env.environment_configuration.render_mode = "human"

    rng = jax.random.PRNGKey(0)
    state = reset_fn(rng)

    i = 0
    while True:
        rng, sub_rng = jax.random.split(rng, 2)
        actions = env.action_space.sample(rng)
        state = step_fn(state=state, action=actions)
        env.render(state=state)

        if state.terminated | state.truncated:
            rng, sub_rng = jax.random.split(rng, 2)
            state = reset_fn(rng)

        i += 1



from typing import Tuple, List

import chex
import cv2
import jax.numpy as jnp
import jax.random
import numpy as np

from biorobot.brittle_star.environment.directed_locomotion.dual import (
    BrittleStarDirectedLocomotionEnvironment,
)
from biorobot.brittle_star.environment.directed_locomotion.shared import (
    BrittleStarDirectedLocomotionEnvironmentConfiguration,
)
from biorobot.brittle_star.mjcf.arena.aquarium import (
    AquariumArenaConfiguration,
    MJCFAquariumArena,
)
from biorobot.brittle_star.mjcf.morphology.morphology import MJCFBrittleStarMorphology
from biorobot.brittle_star.mjcf.morphology.specification.default import (
    default_brittle_star_morphology_specification,
)


def post_render(
    render_output: List[np.ndarray],
    environment_configuration: BrittleStarDirectedLocomotionEnvironmentConfiguration,
) -> None:
    if environment_configuration.render_mode == "human":
        return

    num_cameras = len(environment_configuration.camera_ids)
    num_envs = len(render_output) // num_cameras

    if num_cameras > 1:
        # Horizontally stack frames of the same environment
        frames_per_env = np.array_split(render_output, num_envs)
        render_output = [
            np.concatenate(env_frames, axis=1) for env_frames in frames_per_env
        ]

    # Vertically stack frames of different environments
    stacked_frames = np.concatenate(render_output, axis=0)
    cv2.imshow("render", stacked_frames)
    cv2.waitKey(1)


def create_env(
    backend: str, render_mode: str
) -> BrittleStarDirectedLocomotionEnvironment:
    morphology_spec = default_brittle_star_morphology_specification(
        num_arms=5,
        num_segments_per_arm=5,
        use_p_control=False,
        use_torque_control=True,
        use_tendons=False,
        num_contact_sensors_per_segment=2,
    )
    morphology = MJCFBrittleStarMorphology(morphology_spec)
    arena_config = AquariumArenaConfiguration(attach_target=True)
    arena = MJCFAquariumArena(configuration=arena_config)

    env_config = BrittleStarDirectedLocomotionEnvironmentConfiguration(
        render_mode=render_mode,
        num_physics_steps_per_control_step=10,
        simulation_time=5,
        time_scale=1,
        camera_ids=[0, 1],
        color_contacts=True,
        random_initial_rotation=True,
    )
    env = BrittleStarDirectedLocomotionEnvironment.from_morphology_and_arena(
        morphology=morphology, arena=arena, configuration=env_config, backend=backend
    )
    return env


if __name__ == "__main__":
    BACKEND = "MJC"
    RENDER_MODE = "human"

    env = create_env(backend=BACKEND, render_mode=RENDER_MODE)

    if BACKEND == "MJC":
        env_rng, action_rng = np.random.RandomState(0), None
        step_fn = env.step
        reset_fn = env.reset

        def action_sample_fn(_: None) -> Tuple[np.ndarray, None]:
            return env.action_space.sample(), None

    else:
        env_rng, action_rng = jax.random.split(jax.random.PRNGKey(0), 2)

        step_fn = jax.jit(env.step)
        reset_fn = jax.jit(env.reset)
        reset_fn = env.reset

        def action_sample_fn(rng: chex.PRNGKey) -> Tuple[jnp.ndarray, chex.PRNGKey]:
            rng, sub_rng = jax.random.split(rng, 2)
            return env.action_space.sample(rng=sub_rng), rng

    state = reset_fn(env_rng, [-1.0, 0.0, 0.05])

    while True:
        action, action_rng = action_sample_fn(action_rng)
        state = step_fn(state=state, action=action)
        post_render(env.render(state=state), env.environment_configuration)
        print(state.observations["segment_contact"])
        if state.terminated | state.truncated:
            state = reset_fn(env_rng)
    env.close()



import numpy as np
from moojoco.mjcf.morphology import MJCFMorphology
from transforms3d.euler import euler2quat

from biorobot.brittle_star.mjcf.morphology.parts.arm import MJCFBrittleStarArm
from biorobot.brittle_star.mjcf.morphology.parts.disk import MJCFBrittleStarDisk
from biorobot.brittle_star.mjcf.morphology.specification.default import (
    default_brittle_star_morphology_specification,
)
from biorobot.brittle_star.mjcf.morphology.specification.specification import (
    BrittleStarMorphologySpecification,
)


class MJCFBrittleStarMorphology(MJCFMorphology):
    def __init__(self, specification: BrittleStarMorphologySpecification) -> None:
        super().__init__(specification, name="BrittleStarMorphology")

    @property
    def morphology_specification(self) -> BrittleStarMorphologySpecification:
        return super().morphology_specification

    def _build(self, *args, **kwargs) -> None:
        self._configure_compiler()
        self._configure_defaults()
        self._build_disk()
        self._build_arms()
        self._configure_camera()

    def _configure_compiler(self) -> None:
        self.mjcf_model.compiler.angle = "radian"

    def _configure_defaults(self) -> None:
        self.mjcf_model.default.geom.condim = 4
        self.mjcf_model.default.geom.contype = 1
        self.mjcf_model.default.geom.conaffinity = 0

    def _build_disk(self) -> None:
        self._disk = MJCFBrittleStarDisk(
            parent=self, name="central_disk", pos=np.zeros(3), euler=np.zeros(3)
        )

    def _build_arms(self) -> None:
        # Equally spaced over the disk
        self.arms = []

        disk_radius = self.morphology_specification.disk_specification.radius.value
        arm_angles = np.linspace(0, 2 * np.pi, 6)[:-1]
        number_of_arms = self.morphology_specification.number_of_arms

        for arm_index in range(number_of_arms):
            angle = arm_angles[arm_index]
            position = disk_radius * np.array([np.cos(angle), np.sin(angle), 0])
            arm = MJCFBrittleStarArm(
                parent=self._disk,
                name=f"arm_{arm_index}",
                pos=position,
                euler=[0, 0, angle],
                arm_index=arm_index,
            )
            self.arms.append(arm)

    def _configure_camera(self) -> None:
        self._disk.mjcf_body.add(
            "camera",
            name="side_camera",
            pos=[0.0, -2.0, 2.5],
            quat=euler2quat(40 / 180 * np.pi, 0, 0),
            mode="track",
        )


if __name__ == "__main__":
    spec = default_brittle_star_morphology_specification(
        num_arms=5,
        num_segments_per_arm=5,
        use_p_control=False,
        use_torque_control=True,
        use_tendons=True,
        radius_to_strength_factor=200,
        num_contact_sensors_per_segment=8,
    )
    MJCFBrittleStarMorphology(spec).export_to_xml_with_assets("./mjcf")



from __future__ import annotations

from typing import Union, List

import numpy as np
from dm_control.mjcf.element import _ElementImpl
from moojoco.mjcf.morphology import MJCFMorphology, MJCFMorphologyPart

from biorobot.brittle_star.mjcf.morphology.parts.disk import MJCFBrittleStarDisk
from biorobot.brittle_star.mjcf.morphology.specification.specification import (
    BrittleStarJointSpecification,
    BrittleStarMorphologySpecification,
)
from biorobot.utils import colors
from biorobot.utils.colors import rgba_red, rgba_tendon_relaxed


class MJCFBrittleStarArmSegment(MJCFMorphologyPart):
    def __init__(
        self,
        parent: Union[MJCFMorphology, MJCFMorphologyPart],
        name: str,
        pos: np.array,
        euler: np.array,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(parent, name, pos, euler, *args, **kwargs)

    @property
    def parent(self) -> Union[MJCFBrittleStarDisk, MJCFBrittleStarArmSegment]:
        return super().parent

    @property
    def morphology_specification(self) -> BrittleStarMorphologySpecification:
        return super().morphology_specification

    def _build(self, arm_index: int, segment_index: int) -> None:
        self._arm_index = arm_index
        self._segment_index = segment_index

        self._arm_specification = self.morphology_specification.arm_specifications[
            self._arm_index
        ]
        self._segment_specification = self._arm_specification.segment_specifications[
            self._segment_index
        ]

        self._build_capsule()
        self._build_connector()
        self._configure_joints()
        self._configure_tendons()
        self._configure_actuators()
        self._configure_sensors()

    def _build_capsule(self) -> None:
        radius = self._segment_specification.radius.value
        length = self._segment_specification.length.value

        self._capsule = self.mjcf_body.add(
            "geom",
            name=f"{self.base_name}_capsule",
            type="capsule",
            pos=self.center_of_capsule,
            euler=[0, np.pi / 2, 0],
            size=[radius, length / 2],
            rgba=colors.rgba_green,
            density=self._segment_specification.density.value,
        )

    def _build_connector(self) -> None:
        radius = self._segment_specification.radius.value
        self._connector = self.mjcf_body.add(
            "geom",
            name=f"{self.base_name}_connector",
            type="sphere",
            pos=np.zeros(3),
            size=[0.5 * radius],
            rgba=colors.rgba_gray,
            contype=0,
            conaffinity=0,
            mass=0,
        )

    @property
    def center_of_capsule(self) -> np.ndarray:
        radius = self._segment_specification.radius.value
        length = self._segment_specification.length.value
        x_offset = radius + length / 2
        return np.array([x_offset, 0, 0])

    def _configure_joint(
        self,
        name: str,
        axis: np.ndarray,
        joint_specification: BrittleStarJointSpecification,
    ) -> _ElementImpl:
        joint = self.mjcf_body.add(
            "joint",
            name=name,
            type="hinge",
            limited=True,
            range=[-joint_specification.range.value, joint_specification.range.value],
            axis=axis,
            stiffness=joint_specification.stiffness.value,
            damping=joint_specification.damping.value,
            armature=joint_specification.armature.value,
        )
        return joint

    def _configure_joints(self) -> None:
        self._joints = [
            self._configure_joint(
                name=f"{self.base_name}_in_plane_joint",
                axis=[0, 0, 1],
                joint_specification=self._segment_specification.in_plane_joint_specification,
            ),
            self._configure_joint(
                name=f"{self.base_name}_out_of_plane_joint",
                axis=[0, -1, 0],
                joint_specification=self._segment_specification.out_of_plane_joint_specification,
            ),
        ]

    def _configure_tendon_attachment_points(self) -> None:
        angles = np.linspace(np.pi / 4, 7 * np.pi / 4, 4)
        self._proximal_taps = []
        self.distal_taps = []
        self._connector_taps = []
        for i, angle in enumerate(angles):
            # proximal
            pos = (
                0.8
                * self._segment_specification.radius.value
                * np.array([0, np.cos(angle), np.sin(angle)])
            )
            pos[0] = self._segment_specification.radius.value
            self._proximal_taps.append(
                self.mjcf_body.add(
                    "site",
                    name=f"{self.base_name}_proximal_tap_{i}",
                    type="sphere",
                    rgba=rgba_red,
                    pos=pos,
                    size=[0.001],
                )
            )

            # distal
            pos[0] = (
                self._segment_specification.radius.value
                + self._segment_specification.length.value
            )
            self.distal_taps.append(
                self.mjcf_body.add(
                    "site",
                    name=f"{self.base_name}_distal_tap_{i}",
                    type="sphere",
                    rgba=rgba_red,
                    pos=pos,
                    size=[0.001],
                )
            )

            # Vertebrae
            pos[0] = self._connector.pos[0]
            self._connector_taps.append(
                self.mjcf_body.add(
                    "site",
                    name=f"{self.base_name}_connector_tap_{i}",
                    type="sphere",
                    rgba=rgba_red,
                    pos=pos,
                    size=[0.001],
                )
            )

    def _build_tendons(self) -> None:
        if self._segment_index == 0:
            parent: MJCFBrittleStarDisk = self.parent.parent
            distal_taps = parent.distal_taps[self._arm_index]
        else:
            distal_taps = self.parent.distal_taps

        self._tendons = []
        for tendon_index, (parent_tap, connector_tap, segment_tap) in enumerate(
            zip(distal_taps, self._connector_taps, self._proximal_taps)
        ):
            tendon = self.mjcf_model.tendon.add(
                "spatial",
                name=f"{self.base_name}_tendon_{tendon_index}",
                rgba=rgba_tendon_relaxed,
                width=self._segment_specification.radius.value * 0.1,
            )
            tendon.add("site", site=parent_tap)
            tendon.add("site", site=connector_tap)
            tendon.add("site", site=segment_tap)
            self._tendons.append(tendon)

    def _configure_tendons(self) -> None:
        if self.morphology_specification.actuation_specification.use_tendons.value:
            self._configure_tendon_attachment_points()
            self._build_tendons()

    def _is_first_segment(self) -> bool:
        return self._segment_index == 0

    def _is_last_segment(self) -> bool:
        number_of_segments = len(self._arm_specification.segment_specifications)
        return self._segment_index == number_of_segments - 1

    @property
    def _actuator_strength(self) -> float:
        strength = (
            self._segment_specification.radius.value
            * self.morphology_specification.actuation_specification.radius_to_strength_factor.value
        )
        return strength

    @property
    def _transmissions(self) -> List[_ElementImpl]:
        if self.morphology_specification.actuation_specification.use_tendons.value:
            return self._tendons
        else:
            return self._joints

    def _configure_p_control_actuator(self, transmission: _ElementImpl) -> _ElementImpl:
        actuator_attributes = {
            "name": f"{transmission.name}_p_control",
            "kp": 20,
            "ctrllimited": True,
            "ctrlrange": transmission.range,
            "forcelimited": True,
            "forcerange": [-self._actuator_strength, self._actuator_strength],
            "joint": transmission,
        }

        return self.mjcf_model.actuator.add("position", **actuator_attributes)

    def _configure_p_control_actuators(self) -> None:
        if self.morphology_specification.actuation_specification.use_p_control.value:
            self._actuators = [
                self._configure_p_control_actuator(transmission)
                for transmission in self._transmissions
            ]

    def _configure_torque_control_actuator(
        self, transmission: _ElementImpl
    ) -> _ElementImpl:
        actuator_attributes = {
            "name": f"{transmission.name}_torque_control",
            "ctrllimited": True,
            "forcelimited": True,
            "ctrlrange": [-self._actuator_strength, self._actuator_strength],
            "forcerange": [-self._actuator_strength, self._actuator_strength],
        }

        if self.morphology_specification.actuation_specification.use_tendons.value:
            actuator_attributes["tendon"] = transmission
            actuator_attributes["ctrlrange"] = [-self._actuator_strength, 0]
            gear = 15
            actuator_attributes["gear"] = [gear]
            actuator_attributes["forcerange"] = [-self._actuator_strength * gear, 0]
        else:
            actuator_attributes["joint"] = transmission
            actuator_attributes["ctrlrange"] = [
                -self._actuator_strength,
                self._actuator_strength,
            ]
            actuator_attributes["forcerange"] = [
                -self._actuator_strength,
                self._actuator_strength,
            ]

        return self.mjcf_model.actuator.add("motor", **actuator_attributes)

    def _configure_torque_control_actuators(self) -> None:
        if (
            self.morphology_specification.actuation_specification.use_torque_control.value
        ):
            self._actuators = [
                self._configure_torque_control_actuator(transmission)
                for transmission in self._transmissions
            ]

    def _configure_actuators(self) -> None:
        self._configure_p_control_actuators()
        self._configure_torque_control_actuators()

    def _configure_position_sensor(self) -> None:
        self.mjcf_model.sensor.add(
            "framepos",
            name=f"{self.base_name}_position_sensor",
            objtype="geom",
            objname=self._capsule.name,
        )

    def _configure_joints_sensors(self) -> None:
        for joint in self._joints:
            self.mjcf_model.sensor.add(
                "jointpos", joint=joint, name=f"{joint.name}_jointpos_sensor"
            )
            self.mjcf_model.sensor.add(
                "jointvel", joint=joint, name=f"{joint.name}_jointvel_sensor"
            )
            self.mjcf_model.sensor.add(
                "jointactuatorfrc", joint=joint, name=f"{joint.name}_actuatorfrc_sensor"
            )

    def _configure_actuator_sensors(self) -> None:
        for actuator in self._actuators:
            self.mjcf_model.sensor.add(
                "actuatorfrc",
                actuator=actuator,
                name=f"{actuator.name}_actuatorfrc_sensor",
            )

    def _configure_tendon_sensors(self) -> None:
        if self.morphology_specification.actuation_specification.use_tendons.value:
            for tendon in self._tendons:
                self.mjcf_model.sensor.add(
                    "tendonpos", name=f"{tendon.name}_tendonpos_sensor", tendon=tendon
                )
                self.mjcf_model.sensor.add(
                    "tendonvel", name=f"{tendon.name}_tendonvel_sensor", tendon=tendon
                )

    def _configure_contact_sensors(self) -> None:
        num_contact_sensors = (
            self.morphology_specification.sensor_specification.num_contact_sensors_per_segment.value
        )

        contact_sites = []
        if num_contact_sensors == 1:
            contact_sites.append(
                self.mjcf_body.add(
                    "site",
                    type="capsule",
                    pos=self._capsule.pos,
                    size=self._capsule.size * np.array([1.01, 1]),
                    rgba=rgba_red * np.array([1, 1, 1, 0.5]),
                    euler=self._capsule.euler,
                    name=f"{self.base_name}_contact_site",
                    group=3,
                ),
            )
        else:
            angles = np.linspace(-np.pi / 2, 1.5 * np.pi, num_contact_sensors + 1)[
                :num_contact_sensors
            ]
            radius = self._segment_specification.radius.value

            for i, angle in enumerate(angles):
                pos = self.center_of_capsule + 0.95 * radius * np.array(
                    [0, np.cos(angle), np.sin(angle)]
                )
                contact_sites.append(
                    self.mjcf_body.add(
                        "site",
                        pos=pos,
                        euler=[angle, 0, 0],
                        type="box",
                        size=[
                            self._segment_specification.length.value / 2 + radius / 2,
                            0.05 * radius,
                            0.4 * radius,
                        ],
                        rgba=rgba_red * np.array([1, 1, 1, 0.5]),
                        group=3,
                        name=f"{self.base_name}_contact_site_{i}",
                    )
                )

        for site in contact_sites:
            self.mjcf_model.sensor.add(
                "touch", name=f"{site.name}_contact_sensor", site=site
            )

    def _configure_sensors(self) -> None:
        self._configure_position_sensor()
        self._configure_joints_sensors()
        self._configure_actuator_sensors()
        self._configure_tendon_sensors()
        self._configure_contact_sensors()



from typing import Union

import numpy as np
from moojoco.mjcf.morphology import MJCFMorphology, MJCFMorphologyPart
from scipy.spatial.transform import Rotation as R

from biorobot.brittle_star.mjcf.morphology.specification.specification import (
    BrittleStarMorphologySpecification,
)
from biorobot.utils import colors
from biorobot.utils.colors import rgba_red


class MJCFBrittleStarDisk(MJCFMorphologyPart):
    def __init__(
        self,
        parent: Union[MJCFMorphology, MJCFMorphologyPart],
        name: str,
        pos: np.array,
        euler: np.array,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(parent, name, pos, euler, *args, **kwargs)

    @property
    def morphology_specification(self) -> BrittleStarMorphologySpecification:
        return super().morphology_specification

    def _build(self) -> None:
        self._disk_specification = self.morphology_specification.disk_specification

        self._build_pentagon()
        self._build_arm_connections()
        self._configure_tendon_attachment_points()
        self._configure_sensors()

    def _build_pentagon(self) -> None:
        # Todo: replace this with a dynamically generated mesh
        radius = self.morphology_specification.disk_specification.radius.value
        height = self.morphology_specification.disk_specification.height.value
        # box size
        arm_angle_delta = np.pi / 5
        box_y_size = radius * np.sin(arm_angle_delta)
        box_x_size = radius * np.cos(arm_angle_delta) / 2

        angles = np.linspace(0, 2 * np.pi, 6)[:-1]
        angles += angles[1] / 2

        for i, angle in enumerate(angles):
            pos = box_x_size * np.array([np.cos(angle), np.sin(angle), 0.0])
            self.mjcf_body.add(
                "geom",
                type="box",
                name=f"{self.base_name}_pentagon_side_{i}",
                pos=pos,
                euler=[0, 0, angle],
                size=[box_x_size, box_y_size, height],
                rgba=colors.rgba_green,
                contype=0,
                conaffinity=0,
                mass=0,
            )

        self.mjcf_body.add(
            "geom",
            type="cylinder",
            name=f"{self.base_name}_pentagon_collider",
            pos=[0.0, 0.0, 0.0],
            euler=[0, 0, 0],
            size=[radius, height],
            rgba=[0, 0, 0, 0],
            contype=1,
            conaffinity=0,
            density=self._disk_specification.density.value,
        )

    def _build_arm_connections(self) -> None:
        radius = self.morphology_specification.disk_specification.radius.value
        height = self.morphology_specification.disk_specification.height.value

        arm_angles = np.linspace(0, 2 * np.pi, 6)[:-1]

        connector_length = radius * 0.2
        for i, angle in enumerate(arm_angles):
            pos = (radius - connector_length) * np.array(
                [np.cos(angle), np.sin(angle), 0.0]
            )
            self.mjcf_body.add(
                "geom",
                type="box",
                name=f"{self.base_name}_arm_connector_{i}",
                pos=pos,
                euler=[0.0, 0.0, angle],
                size=[connector_length, height, height],
                rgba=colors.rgba_green,
                contype=0,
                conaffinity=0,
            )

    def _configure_tendon_attachment_points(self) -> None:
        if self.morphology_specification.actuation_specification.use_tendons.value:
            self.distal_taps = []

            disk_radius = self.morphology_specification.disk_specification.radius.value
            center_pos = np.array([disk_radius, 0, 0])

            arm_angles = np.linspace(0, 2 * np.pi, 6)[:-1]
            tap_angles = np.linspace(np.pi / 4, 7 * np.pi / 4, 4)

            for arm_index, arm_angle in enumerate(arm_angles):
                arm_specification = self.morphology_specification.arm_specifications[
                    arm_index
                ]
                if arm_specification.number_of_segments == 0:
                    continue

                base_segment_radius = arm_specification.segment_specifications[
                    0
                ].radius.value

                arm_taps = []
                positions = []
                for angle in tap_angles:
                    pos = 0.8 * center_pos + 0.8 * base_segment_radius * np.array(
                        [0, np.cos(angle), np.sin(angle)]
                    )
                    positions.append(pos)

                for tap_index, position in enumerate(positions):
                    # rotate position around arm_angle degress
                    # Define the rotation
                    rotation = R.from_euler("z", arm_angle, degrees=False)

                    # Rotate point A around point B
                    rotated_point = rotation.apply(position)

                    arm_taps.append(
                        self.mjcf_body.add(
                            "site",
                            name=f"{self.base_name}_arm_{arm_index}_tap_{tap_index}",
                            type="sphere",
                            rgba=rgba_red,
                            pos=rotated_point,
                            size=[0.001],
                        )
                    )
                self.distal_taps.append(arm_taps)

    def _configure_sensors(self) -> None:
        self.mjcf_model.sensor.add(
            "framepos",
            name=f"{self.base_name}_framepos_sensor",
            objtype="xbody",
            objname=self.mjcf_body.name,
        )
        self.mjcf_model.sensor.add(
            "framequat",
            name=f"{self.base_name}_framequat_sensor",
            objtype="xbody",
            objname=self.mjcf_body.name,
        )
        self.mjcf_model.sensor.add(
            "framelinvel",
            name=f"{self.base_name}_framelinvel_sensor",
            objtype="xbody",
            objname=self.mjcf_body.name,
        )
        self.mjcf_model.sensor.add(
            "frameangvel",
            name=f"{self.base_name}_frameangvel_sensor",
            objtype="xbody",
            objname=self.mjcf_body.name,
        )
